/*
 *  cjoboard.h
 *  Othello
 *
 *  Created by Curtis Jones on 2010.09.27.
 *  Copyright 2010 __MyCompanyName__. All rights reserved.
 *
 */

#ifndef __CJO_BOARD_H__
#define __CJO_BOARD_H__

#include <stdint.h>
#include "cjospace.h"

//
// returns the state value of the given position (y) for the given board (x)
//
#define CJO_BOARD_GET_STATE(x,y) ((x)->board[CJO_SPACE_POS(y)])

//
//
//
#define CJO_BOARD_SET_STATE(x,y,z) { \
	if (CJO_STATE_WHITE == z) { \
		if (CJO_BOARD_GET_STATE(x,y) == CJO_STATE_BLACK) \
			(x)->black -= 1; \
		(x)->white += 1; \
	} \
	else if (CJO_STATE_BLACK == z) { \
		if (CJO_BOARD_GET_STATE(x,y) == CJO_STATE_WHITE) \
			(x)->white -= 1; \
		(x)->black += 1; \
	} \
	else if (CJO_STATE_NULL == z) { \
		if (CJO_BOARD_GET_STATE(x,y) == CJO_STATE_BLACK) \
			(x)->black -= 1; \
		else if (CJO_BOARD_GET_STATE(x,y) == CJO_STATE_WHITE) \
			(x)->white -= 1; \
	} \
	(x)->board[CJO_SPACE_POS(y)] = (z); \
}

//
// cjoboard
//
struct cjoboard
{
	uint32_t board[64];                  // sixty-four spaces of cjostate_t values
	uint8_t white;                       // number of white pieces on the board
	uint8_t black;                       // number of black pieces on the board
};
typedef struct cjoboard cjoboard_t;

/**
 * Initialize a new board.
 */
int cjoboard_init (cjoboard_t*);

/**
 * Make a play on the board, at the space, with the state. Returns the number of pieces taken in the
 * play; or a negative value if the play is illegal or an error occurred.
 */
int cjoboard_play (cjoboard_t*, cjospace_t, cjostate_t);

/**
 * Returns the number of plays available on the board for the given user.
 */
int cjoboard_playcnt (cjoboard_t*, cjostate_t);

/**
 * Returns the number of plays available on the board for the given user and add information about
 * each play to the plays struct.
 */
int cjoboard_playlst (cjoboard_t*, cjostate_t, cjoplays_t*);

/**
 * Print an ascii representation of the board to stdout.
 */
void cjoboard_print (cjoboard_t*);

#endif /* __CJO_BOARD_H__ */
