/*
 *  cjospace.h
 *  Othello
 *
 *  Created by Curtis Jones on 2010.09.27.
 *  Copyright 2010 __MyCompanyName__. All rights reserved.
 *
 */

#ifndef __CJO_SPACE_H__
#define __CJO_SPACE_H__

#include <assert.h>
#include <stdint.h>

//
// this mask is used to get the row-only value out of a cjospace_t value
//
extern const uint32_t CJO_SPACE_ROW_MASK;

//
// this mask is used to get the column-only value out of a cjospace_t value
//
extern const uint32_t CJO_SPACE_COL_MASK;

//
// this mask is used to get the ring-only value out of a cjospace_t value
//
extern const uint32_t CJO_SPACE_RING_MASK;

//
// this mask is used to get the position-only value of a cjospace_t value
//
extern const uint32_t CJO_SPACE_POS_MASK;

//
// cjostate_t
//
typedef enum
{
	CJO_STATE_WHITE = 0,
	CJO_STATE_BLACK = 1,
	CJO_STATE_NULL  = 2
} cjostate_t;

//
// cjodir_t
//
typedef enum
{
	CJO_DIR_N  = 0x01,
	CJO_DIR_S  = 0x02,
	CJO_DIR_W  = 0x04,
	CJO_DIR_E  = 0x08,
	CJO_DIR_NW = 0x10,
	CJO_DIR_NE = 0x20,
	CJO_DIR_SW = 0x40,
	CJO_DIR_SE = 0x80
} cjodir_t;

//
// cjorow_t
//
typedef enum
{
	CJO_SPACE_ROW_A = 0x0000,
	CJO_SPACE_ROW_B = 0x1000,
	CJO_SPACE_ROW_C = 0x2000,
	CJO_SPACE_ROW_D = 0x3000,
	CJO_SPACE_ROW_E = 0x4000,
	CJO_SPACE_ROW_F = 0x5000,
	CJO_SPACE_ROW_G = 0x6000,
	CJO_SPACE_ROW_H = 0x7000
} cjorow_t;

//
// cjocol_t
//
typedef enum
{
	CJO_SPACE_COL_1 = 0x0000,
	CJO_SPACE_COL_2 = 0x0100,
	CJO_SPACE_COL_3 = 0x0200,
	CJO_SPACE_COL_4 = 0x0300,
	CJO_SPACE_COL_5 = 0x0400,
	CJO_SPACE_COL_6 = 0x0500,
	CJO_SPACE_COL_7 = 0x0600,
	CJO_SPACE_COL_8 = 0x0700
} cjocol_t;

//
// cjoring_t
//
typedef enum
{
	CJO_SPACE_RING_A = 0x00000,
	CJO_SPACE_RING_B = 0x10000,
	CJO_SPACE_RING_C = 0x20000,
	CJO_SPACE_RING_D = 0x30000
} cjoring_t;

//
// cjospace_t
//
// the high byte is used to store the row and column while the low byte is used to store the 
// position offset within the board (to avoid repetitive multiplication of the row/col). the four
// high-bits of the high byte store the row, while the four low-bits of the high bytes store the
// column.
//
typedef enum
{
	CJO_SPACE_A1 = CJO_SPACE_RING_A | CJO_SPACE_ROW_A | CJO_SPACE_COL_1 |  0,
	CJO_SPACE_A2 = CJO_SPACE_RING_A | CJO_SPACE_ROW_A | CJO_SPACE_COL_2 |  1,
	CJO_SPACE_A3 = CJO_SPACE_RING_A | CJO_SPACE_ROW_A | CJO_SPACE_COL_3 |  2,
	CJO_SPACE_A4 = CJO_SPACE_RING_A | CJO_SPACE_ROW_A | CJO_SPACE_COL_4 |  3,
	CJO_SPACE_A5 = CJO_SPACE_RING_A | CJO_SPACE_ROW_A | CJO_SPACE_COL_5 |  4,
	CJO_SPACE_A6 = CJO_SPACE_RING_A | CJO_SPACE_ROW_A | CJO_SPACE_COL_6 |  5,
	CJO_SPACE_A7 = CJO_SPACE_RING_A | CJO_SPACE_ROW_A | CJO_SPACE_COL_7 |  6,
	CJO_SPACE_A8 = CJO_SPACE_RING_A | CJO_SPACE_ROW_A | CJO_SPACE_COL_8 |  7,
	
	CJO_SPACE_B1 = CJO_SPACE_RING_A | CJO_SPACE_ROW_B | CJO_SPACE_COL_1 |  8,
	CJO_SPACE_B2 = CJO_SPACE_RING_B | CJO_SPACE_ROW_B | CJO_SPACE_COL_2 |  9,
	CJO_SPACE_B3 = CJO_SPACE_RING_B | CJO_SPACE_ROW_B | CJO_SPACE_COL_3 | 10,
	CJO_SPACE_B4 = CJO_SPACE_RING_B | CJO_SPACE_ROW_B | CJO_SPACE_COL_4 | 11,
	CJO_SPACE_B5 = CJO_SPACE_RING_B | CJO_SPACE_ROW_B | CJO_SPACE_COL_5 | 12,
	CJO_SPACE_B6 = CJO_SPACE_RING_B | CJO_SPACE_ROW_B | CJO_SPACE_COL_6 | 13,
	CJO_SPACE_B7 = CJO_SPACE_RING_B | CJO_SPACE_ROW_B | CJO_SPACE_COL_7 | 14,
	CJO_SPACE_B8 = CJO_SPACE_RING_A | CJO_SPACE_ROW_B | CJO_SPACE_COL_8 | 15,
	
	CJO_SPACE_C1 = CJO_SPACE_RING_A | CJO_SPACE_ROW_C | CJO_SPACE_COL_1 | 16,
	CJO_SPACE_C2 = CJO_SPACE_RING_B | CJO_SPACE_ROW_C | CJO_SPACE_COL_2 | 17,
	CJO_SPACE_C3 = CJO_SPACE_RING_C | CJO_SPACE_ROW_C | CJO_SPACE_COL_3 | 18,
	CJO_SPACE_C4 = CJO_SPACE_RING_C | CJO_SPACE_ROW_C | CJO_SPACE_COL_4 | 19,
	CJO_SPACE_C5 = CJO_SPACE_RING_C | CJO_SPACE_ROW_C | CJO_SPACE_COL_5 | 20,
	CJO_SPACE_C6 = CJO_SPACE_RING_C | CJO_SPACE_ROW_C | CJO_SPACE_COL_6 | 21,
	CJO_SPACE_C7 = CJO_SPACE_RING_B | CJO_SPACE_ROW_C | CJO_SPACE_COL_7 | 22,
	CJO_SPACE_C8 = CJO_SPACE_RING_A | CJO_SPACE_ROW_C | CJO_SPACE_COL_8 | 23,
	
	CJO_SPACE_D1 = CJO_SPACE_RING_A | CJO_SPACE_ROW_D | CJO_SPACE_COL_1 | 24,
	CJO_SPACE_D2 = CJO_SPACE_RING_B | CJO_SPACE_ROW_D | CJO_SPACE_COL_2 | 25,
	CJO_SPACE_D3 = CJO_SPACE_RING_C | CJO_SPACE_ROW_D | CJO_SPACE_COL_3 | 26,
	CJO_SPACE_D4 = CJO_SPACE_RING_D | CJO_SPACE_ROW_D | CJO_SPACE_COL_4 | 27,
	CJO_SPACE_D5 = CJO_SPACE_RING_D | CJO_SPACE_ROW_D | CJO_SPACE_COL_5 | 28,
	CJO_SPACE_D6 = CJO_SPACE_RING_C | CJO_SPACE_ROW_D | CJO_SPACE_COL_6 | 29,
	CJO_SPACE_D7 = CJO_SPACE_RING_B | CJO_SPACE_ROW_D | CJO_SPACE_COL_7 | 30,
	CJO_SPACE_D8 = CJO_SPACE_RING_A | CJO_SPACE_ROW_D | CJO_SPACE_COL_8 | 31,
	
	CJO_SPACE_E1 = CJO_SPACE_RING_A | CJO_SPACE_ROW_E | CJO_SPACE_COL_1 | 32,
	CJO_SPACE_E2 = CJO_SPACE_RING_B | CJO_SPACE_ROW_E | CJO_SPACE_COL_2 | 33,
	CJO_SPACE_E3 = CJO_SPACE_RING_C | CJO_SPACE_ROW_E | CJO_SPACE_COL_3 | 34,
	CJO_SPACE_E4 = CJO_SPACE_RING_D | CJO_SPACE_ROW_E | CJO_SPACE_COL_4 | 35,
	CJO_SPACE_E5 = CJO_SPACE_RING_D | CJO_SPACE_ROW_E | CJO_SPACE_COL_5 | 36,
	CJO_SPACE_E6 = CJO_SPACE_RING_C | CJO_SPACE_ROW_E | CJO_SPACE_COL_6 | 37,
	CJO_SPACE_E7 = CJO_SPACE_RING_B | CJO_SPACE_ROW_E | CJO_SPACE_COL_7 | 38,
	CJO_SPACE_E8 = CJO_SPACE_RING_A | CJO_SPACE_ROW_E | CJO_SPACE_COL_8 | 39,
	
	CJO_SPACE_F1 = CJO_SPACE_RING_A | CJO_SPACE_ROW_F | CJO_SPACE_COL_1 | 40,
	CJO_SPACE_F2 = CJO_SPACE_RING_B | CJO_SPACE_ROW_F | CJO_SPACE_COL_2 | 41,
	CJO_SPACE_F3 = CJO_SPACE_RING_C | CJO_SPACE_ROW_F | CJO_SPACE_COL_3 | 42,
	CJO_SPACE_F4 = CJO_SPACE_RING_C | CJO_SPACE_ROW_F | CJO_SPACE_COL_4 | 43,
	CJO_SPACE_F5 = CJO_SPACE_RING_C | CJO_SPACE_ROW_F | CJO_SPACE_COL_5 | 44,
	CJO_SPACE_F6 = CJO_SPACE_RING_C | CJO_SPACE_ROW_F | CJO_SPACE_COL_6 | 45,
	CJO_SPACE_F7 = CJO_SPACE_RING_B | CJO_SPACE_ROW_F | CJO_SPACE_COL_7 | 46,
	CJO_SPACE_F8 = CJO_SPACE_RING_A | CJO_SPACE_ROW_F | CJO_SPACE_COL_8 | 47,
	
	CJO_SPACE_G1 = CJO_SPACE_RING_A | CJO_SPACE_ROW_G | CJO_SPACE_COL_1 | 48,
	CJO_SPACE_G2 = CJO_SPACE_RING_B | CJO_SPACE_ROW_G | CJO_SPACE_COL_2 | 49,
	CJO_SPACE_G3 = CJO_SPACE_RING_B | CJO_SPACE_ROW_G | CJO_SPACE_COL_3 | 50,
	CJO_SPACE_G4 = CJO_SPACE_RING_B | CJO_SPACE_ROW_G | CJO_SPACE_COL_4 | 51,
	CJO_SPACE_G5 = CJO_SPACE_RING_B | CJO_SPACE_ROW_G | CJO_SPACE_COL_5 | 52,
	CJO_SPACE_G6 = CJO_SPACE_RING_B | CJO_SPACE_ROW_G | CJO_SPACE_COL_6 | 53,
	CJO_SPACE_G7 = CJO_SPACE_RING_B | CJO_SPACE_ROW_G | CJO_SPACE_COL_7 | 54,
	CJO_SPACE_G8 = CJO_SPACE_RING_A | CJO_SPACE_ROW_G | CJO_SPACE_COL_8 | 55,
	
	CJO_SPACE_H1 = CJO_SPACE_RING_A | CJO_SPACE_ROW_H | CJO_SPACE_COL_1 | 56,
	CJO_SPACE_H2 = CJO_SPACE_RING_A | CJO_SPACE_ROW_H | CJO_SPACE_COL_2 | 57,
	CJO_SPACE_H3 = CJO_SPACE_RING_A | CJO_SPACE_ROW_H | CJO_SPACE_COL_3 | 58,
	CJO_SPACE_H4 = CJO_SPACE_RING_A | CJO_SPACE_ROW_H | CJO_SPACE_COL_4 | 59,
	CJO_SPACE_H5 = CJO_SPACE_RING_A | CJO_SPACE_ROW_H | CJO_SPACE_COL_5 | 60,
	CJO_SPACE_H6 = CJO_SPACE_RING_A | CJO_SPACE_ROW_H | CJO_SPACE_COL_6 | 61,
	CJO_SPACE_H7 = CJO_SPACE_RING_A | CJO_SPACE_ROW_H | CJO_SPACE_COL_7 | 62,
	CJO_SPACE_H8 = CJO_SPACE_RING_A | CJO_SPACE_ROW_H | CJO_SPACE_COL_8 | 63
	
} cjospace_t;

//
// cjolookup_t
//
typedef struct
{
	cjospace_t val;
} cjolookup_t;

//
// cjoplay_t
//
typedef struct
{
	cjospace_t space;
	uint32_t directions;
	uint32_t wins;
	uint32_t score;
} cjoplay_t;

//
// cjoplays_t
//
typedef struct
{
	uint32_t playcnt;
	uint32_t wins;
	cjoplay_t plays[64];
} cjoplays_t;

//
// this struct maps a space position (0..63) to the full space value (CJO_SPACE_XY)
//
extern cjolookup_t cjo_pos_2_space[];

//
//
//
#define CJO_SPACE_ROW(x) (CJO_SPACE_ROW_MASK & (x))
#define CJO_SPACE_COL(x) (CJO_SPACE_COL_MASK & (x))
#define CJO_SPACE_POS(x) (CJO_SPACE_POS_MASK & (x))
#define CJO_SPACE_RING(x) (CJO_SPACE_RING_MASK & (x))

//
// row comparisons: less than (lt), greater than (gt), equal to (eq), not equal to (ne).
//
#define CJO_SPACE_ROW_LT(x,y) (CJO_SPACE_ROW(x)  < CJO_SPACE_ROW(y))
#define CJO_SPACE_ROW_GT(x,y) (CJO_SPACE_ROW(x)  > CJO_SPACE_ROW(y))
#define CJO_SPACE_ROW_EQ(x,y) (CJO_SPACE_ROW(x) == CJO_SPACE_ROW(y))
#define CJO_SPACE_ROW_NE(x,y) (CJO_SPACE_ROW(x) != CJO_SPACE_ROW(y))
#define CJO_SPACE_ROW_LE(x,y) (CJO_SPACE_ROW(x) <= CJO_SPACE_ROW(y))
#define CJO_SPACE_ROW_GE(x,y) (CJO_SPACE_ROW(x) >= CJO_SPACE_ROW(y))

//
// column comparisons: less than (lt), greater than (gt), equal to (eq), not equal to (ne).
//
#define CJO_SPACE_COL_LT(x,y) (CJO_SPACE_COL(x)  < CJO_SPACE_COL(y))
#define CJO_SPACE_COL_GT(x,y) (CJO_SPACE_COL(x)  > CJO_SPACE_COL(y))
#define CJO_SPACE_COL_EQ(x,y) (CJO_SPACE_COL(x) == CJO_SPACE_COL(y))
#define CJO_SPACE_COL_NE(x,y) (CJO_SPACE_COL(x) != CJO_SPACE_COL(y))
#define CJO_SPACE_COL_LE(x,y) (CJO_SPACE_COL(x) <= CJO_SPACE_COL(y))
#define CJO_SPACE_COL_GE(x,y) (CJO_SPACE_COL(x) >= CJO_SPACE_COL(y))

//
// evalutas to the CJO_SPACE_RING_* associated with the given space.
//
#define CJO_SPACE_CALC_RING(x) (CJO_SPACE_RING(cjo_pos_2_space[CJO_SPACE_POS(x)].val))

//
// returns the opposite player state and null if null is given.
//
// CJO_XSTATE(CJO_STATE_BLACK) = CJO_STATE_WHITE
// CJO_XSTATE(CJO_STATE_WHITE) = CJO_STATE_BLACK
// CJO_XSTATE(CJO_STATE_NULL ) = CJO_STATE_NULL
//
#define CJO_XSTATE(x) ((x) == CJO_STATE_NULL ? CJO_STATE_NULL : (x) == CJO_STATE_BLACK ? CJO_STATE_WHITE : CJO_STATE_BLACK)

//
// null = ' ', black = 'x', white = 'o'
//
#define CJO_STATE_CHAR(x) ((x) == CJO_STATE_NULL ? ' ' : (x) == CJO_STATE_BLACK ? 'x' : 'o')

//
// moves the space north one row. first check that we are not already on the top row. leave the
// column value alone. subtract eight from the position offset.
//
#define CJO_SPACE_MOVE_N(x) { \
	assert(CJO_SPACE_ROW(x) != CJO_SPACE_ROW_A); \
	x = (CJO_SPACE_ROW(x)-0x1000) | CJO_SPACE_COL(x) | (CJO_SPACE_POS(x)-8); \
	x = x | CJO_SPACE_CALC_RING(x); \
}

//
//
//
#define CJO_SPACE_MOVE_S(x) { \
	assert(CJO_SPACE_ROW(x) != CJO_SPACE_ROW_H); \
	x = (CJO_SPACE_ROW(x)+0x1000) | CJO_SPACE_COL(x) | (CJO_SPACE_POS(x)+8); \
	x = x | CJO_SPACE_CALC_RING(x); \
}

//
//
//
#define CJO_SPACE_MOVE_W(x) { \
	assert(CJO_SPACE_COL(x) != CJO_SPACE_COL_1); \
	x = CJO_SPACE_ROW(x) | (CJO_SPACE_COL(x)-0x0100) | (CJO_SPACE_POS(x)-1); \
	x = x | CJO_SPACE_CALC_RING(x); \
}

//
//
//
#define CJO_SPACE_MOVE_E(x) { \
	assert(CJO_SPACE_COL(x) != CJO_SPACE_COL_8); \
	x = CJO_SPACE_ROW(x) | (CJO_SPACE_COL(x)+0x0100) | (CJO_SPACE_POS(x)+1); \
	x = x | CJO_SPACE_CALC_RING(x); \
}

//
//
//
#define CJO_SPACE_MOVE_NW(x) { \
	CJO_SPACE_MOVE_N(x) \
	CJO_SPACE_MOVE_W(x) \
}

//
//
//
#define CJO_SPACE_MOVE_NE(x) { \
	CJO_SPACE_MOVE_N(x) \
	CJO_SPACE_MOVE_E(x) \
}

//
//
//
#define CJO_SPACE_MOVE_SW(x) { \
	CJO_SPACE_MOVE_S(x) \
	CJO_SPACE_MOVE_W(x) \
}

//
//
//
#define CJO_SPACE_MOVE_SE(x) { \
	CJO_SPACE_MOVE_S(x) \
	CJO_SPACE_MOVE_E(x) \
}

#endif /* __CJO_BOARD_H__ */
